@@ -19,11 +19,17 @@ log = Logger()
 
 from xpra.protocol import Protocol
 from xpra.keys import mask_to_names, MODIFIER_NAMES
-from xpra.platform.gui import ClipboardProtocolHelper, ClientExtras, grok_modifier_map
+from xpra.platform.gui import ClientExtras
+from xpra.scripts.main import ENCODINGS
 
 import xpra
 default_capabilities = {"__prerelease_version": xpra.__version__}
 
+def nn(x):
+    if x is None:
+        return  ""
+    return x
+
 class ClientSource(object):
     def __init__(self, protocol):
         self._ordinary_packets = []
@@ -172,8 +178,13 @@ class ClientWindow(gtk.Window):
 
     def refresh_window(self):
         log.debug("Automatic refresh for id ", self._id)
-        self._client.send(["buffer-refresh", self._id, True, 95])
-        self._refresh_requested = True
+        self._client.send_refresh(self, self._id)
+
+    def refresh_all_windows(self):
+        #this method is only here because we may want to fire it
+        #from a --key-shortcut action and the event is delivered to
+        #the "ClientWindow"
+        self._client.send_refresh_all()
 
     def draw(self, x, y, width, height, coding, img_data):
         gc = self._backing.new_gc()
@@ -184,22 +195,7 @@ class ClientWindow(gtk.Window):
             pixbuf = loader.get_pixbuf()
             if not pixbuf:
                 if self._failed_pixbuf_index<10:
-                    import os.path, sys
-                    if sys.platform.startswith("win"):
-                        appdata = os.environ.get("APPDATA")
-                        if not os.path.exists(appdata):
-                            os.mkdir(appdata)
-                        xpra_path = os.path.join(appdata, "Xpra")
-                        if not os.path.exists(xpra_path):
-                            os.mkdir(xpra_path)
-                    else:
-                        xpra_path = os.path.expanduser("~/.xpra")
-                    failed_pixbuf_file = os.path.join(xpra_path, "failed-%s.%s" % (self._failed_pixbuf_index, coding))
-                    f = open(failed_pixbuf_file, 'wb')
-                    f.write(img_data)
-                    f.close()
-                    self._failed_pixbuf_index += 1
-                    log.error("failed %s pixbuf=%s data saved to %s, len=%s" % (coding, pixbuf, failed_pixbuf_file, len(img_data)))
+                    log.error("failed %s pixbuf=%s data len=%s" % (coding, pixbuf, len(img_data)))
                 elif self._failed_pixbuf_index==10:
                     log.error("too many pixbuf failures! (will no longer be logged)")
                     self._failed_pixbuf_index += 1
@@ -275,56 +271,11 @@ class ClientWindow(gtk.Window):
     def void(self):
         pass
 
-    def _key_action(self, event, depressed):
-        log.debug("key_action(%s,%s)" % (event, depressed))
-        modifiers = self._client.mask_to_names(event.state)
-        name = gtk.gdk.keyval_name(event.keyval)
-        shortcut = self._client.key_shortcuts.get(name)
-        if shortcut:
-            (req_mods, action) = shortcut
-            mods_found = True
-            for rm in req_mods:
-                if rm not in modifiers:
-                    mods_found = False
-                    break
-            if mods_found:
-                if not depressed:
-                    """ when the key is released, just ignore it - do NOT send it to the server! """
-                    return
-                try:
-                    method = getattr(self, action)
-                    log.info("key_action(%s,%s) has been handled by shortcut=%s" % (event, depressed, shortcut))
-                except AttributeError, e:
-                    log.error("key dropped, invalid method name in shortcut %s: %s" % (action, e))
-                    return
-                try:
-                    method()
-                    return
-                except Exception, e:
-                    log.error("key_action(%s,%s) failed to execute shortcut=%s: %s" % (event, depressed, shortcut, e))
-        
-        def nn(arg):
-            if arg is not None:
-                return  arg
-            return  ""
-        if self._client._raw_keycodes_feature:
-            """ for versions newer than 0.0.7.24, we send ALL the raw information we have """
-            keycode = event.hardware_keycode
-            log.debug("key_action(%s,%s) modifiers=%s, name=%s, state=%s, keyval=%s, string=%s, keycode=%s" % (event, depressed, modifiers, name, event.state, event.keyval, event.string, keycode))
-            self._client.send(["key-action", self._id, nn(name), depressed, modifiers, nn(event.keyval), nn(event.string), nn(keycode)])
-        else:
-            """ versions before 0.0.7.24 only accept 4 parameters (no keyval, keycode, ...) """
-            # Apparently some weird keys (e.g. "media keys") can have no keyval or
-            # no keyval name (I believe that both give us a None here).  Another
-            # reason to upgrade to the version above
-            if name is not None:
-                self._client.send(["key-action", self._id, nn(name), depressed, modifiers])
-
     def do_key_press_event(self, event):
-        self._key_action(event, True)
+        self._client.handle_key_action(event, self, True)
 
     def do_key_release_event(self, event):
-        self._key_action(event, False)
+        self._client.handle_key_action(event, self, False)
 
     def _pointer_modifiers(self, event):
         pointer = (int(event.x_root), int(event.y_root))
@@ -368,6 +319,7 @@ gobject.type_register(ClientWindow)
 
 class XpraClient(gobject.GObject):
     __gsignals__ = {
+        "clipboard-toggled": n_arg_signal(0),
         "handshake-complete": n_arg_signal(0),
         "received-gibberish": n_arg_signal(1),
         }
@@ -380,8 +332,10 @@ class XpraClient(gobject.GObject):
         if opts.title_suffix is not None:
             title = "@title@ %s" % opts.title_suffix
         self.title = title
+        self.session_name = opts.session_name
         self.password_file = opts.password_file
         self.compression_level = opts.compression_level
+        self.encoding = opts.encoding
         self.jpegquality = opts.jpegquality
         self.auto_refresh_delay = opts.auto_refresh_delay
         self.max_bandwidth = opts.max_bandwidth
@@ -390,9 +344,18 @@ class XpraClient(gobject.GObject):
             """ jpegquality was not set, use a better start value """
             self.jpegquality = 50
 
+        self.server_capabilities = {}
+        self.bell_enabled = True
+        self.notifications_enabled = True
+        self._client_extras = ClientExtras(self, opts)
+        self.clipboard_enabled = opts.clipboard and self._client_extras.supports_clipboard()
+
         self._protocol = Protocol(conn, self.process_packet)
         ClientSource(self._protocol)
 
+        self.key_repeat_delay = -1
+        self.key_repeat_interval = -1
+        self.keycodes_pressed = {}
         self._raw_keycodes_feature = False
         self._focus_modifiers_feature = False
         self._remote_version = None
@@ -405,30 +368,24 @@ class XpraClient(gobject.GObject):
         self._xsettings_watcher = None
         self._root_props_watcher = None
 
-        # FIXME: these should perhaps be merged.
-        if opts.clipboard:
-            self._clipboard_helper = ClipboardProtocolHelper(self.send)
-        else:
-            self._clipboard_helper = None
-        self._client_extras = ClientExtras(self.send, opts.pulseaudio, opts)
-
         self._focused = None
         def compute_receive_bandwidth(delay):
             bw = (self._protocol._recv_counter / 1024) * 1000/ delay;
             self._protocol._recv_counter = 0;
             log.debug("Bandwidth is ", bw, "kB/s, max ", self.max_bandwidth, "kB/s")
 
+            q = self.jpegquality
             if bw > self.max_bandwidth:
-                self.jpegquality -= 10;
+                q -= 10;
             elif bw < self.max_bandwidth:
-                self.jpegquality += 5;
+                q += 5;
 
-            if self.jpegquality > 95:
-                self.jpegquality = 95;
-            elif self.jpegquality < 10:
-                self.jpegquality = 10;
+            if q > 95:
+                q = 95;
+            elif q < 10:
+                q = 10;
 
-            self.send_jpeg_quality()
+            self.send_jpeg_quality(q)
             return True
 
         if (self.max_bandwidth):
@@ -438,6 +395,7 @@ class XpraClient(gobject.GObject):
     def run(self):
         gtk_main_quit_on_fatal_exceptions_enable()
         gtk.main()
+        self._client_extras.exit()
         if self._protocol:
             self._protocol.close()
 
@@ -478,37 +436,100 @@ class XpraClient(gobject.GObject):
                 if not valid:
                     continue
             keyname = keyspec[len(keyspec)-1]
-            shortcuts[keyname] = (modifiers, action)                
+            shortcuts[keyname] = (modifiers, action)
         log.debug("parse_shortcuts(%s)=%s" % (str(strs), shortcuts))
         return  shortcuts
 
-    def query_xkbmap(self):
-        from xpra.platform import X11_KEYMAPS
-        if not X11_KEYMAPS:
-            log.info("this platform does not support X11 keymaps, not querying the system with setxkbmap or xmodmap");
-            self.xkbmap_print, self.xkbmap_query, self.xkbmap_query, self.xmodmap_data = None, None, None, None
+    def key_handled_as_shortcut(self, window, key_name, modifiers, depressed):
+        shortcut = self.key_shortcuts.get(key_name)
+        if not shortcut:
+            return  False
+        (req_mods, action) = shortcut
+        for rm in req_mods:
+            if rm not in modifiers:
+                #modifier is missing, bail out
+                return False
+        if not depressed:
+            """ when the key is released, just ignore it - do NOT send it to the server! """
+            return  True
+        try:
+            method = getattr(window, action)
+            log.info("key_handled_as_shortcut(%s,%s,%s,%s) has been handled by shortcut=%s", window, key_name, modifiers, depressed, shortcut)
+        except AttributeError, e:
+            log.error("key dropped, invalid method name in shortcut %s: %s", action, e)
+            return  True
+        try:
+            method()
+        except Exception, e:
+            log.error("key_handled_as_shortcut(%s,%s,%s,%s) failed to execute shortcut=%s: %s", window, key_name, modifiers, depressed, shortcut, e)
+        return  True
+
+    def handle_key_action(self, event, window, depressed):
+        log.debug("handle_key_action(%s,%s,%s)" % (event, window, depressed))
+        modifiers = self.mask_to_names(event.state)
+        name = gtk.gdk.keyval_name(event.keyval)
+        if self.key_handled_as_shortcut(window, name, modifiers, depressed):
             return
-        def get_keyboard_data(command, arg):
-            # Find the client's current keymap so we can send it to the server:
-            try:
-                import subprocess
-                cmd = [command, arg]
-                process = subprocess.Popen(cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=False)
-                (out,_) = process.communicate(None)
-                if process.returncode==0:
-                    return out
-                log.error("'%s %s' failed with exit code %s\n" % (command, arg, process.returncode))
-            except Exception, e:
-                log.error("error running '%s %s': %s\n" % (command, arg, e))
-            return None
-        self.xkbmap_print = get_keyboard_data("setxkbmap", "-print")
-        if self.xkbmap_print is None:
-            log.error("your keyboard mapping will probably be incorrect unless you are using a 'us' layout");
-        self.xkbmap_query = get_keyboard_data("setxkbmap", "-query")
-        if self.xkbmap_query is None and self.xkbmap_print is not None:
-            log.error("the server will try to guess your keyboard mapping, which works reasonably well in most cases");
-            log.error("however, upgrading 'setxkbmap' to a version that supports the '-query' parameter is preferred");
-        self.xmodmap_data = get_keyboard_data("xmodmap", "-pke");
+        id = self._window_to_id[window]
+        if not self._raw_keycodes_feature:
+            """ versions before 0.0.7.24 only accept 4 parameters (no keyval, keycode, ...)
+                also used on win32 and osx since those don't have keymaps/keycode
+            """
+            # Apparently some weird keys (e.g. "media keys") can have no keyval or
+            # no keyval name (I believe that both give us a None here).
+            # Another reason to use the _raw_keycodes_feature wherever possible.
+            if name is not None:
+                self.send(["key-action", id, name, depressed, modifiers])
+            return
+        keycode = event.hardware_keycode
+        log.debug("key_action(%s,%s,%s) modifiers=%s, name=%s, state=%s, keyval=%s, string=%s, keycode=%s" % (event, window, depressed, modifiers, name, event.state, event.keyval, event.string, keycode))
+        self.send(["key-action", id, nn(name), depressed, modifiers, nn(event.keyval), nn(event.string), nn(keycode)])
+        if keycode and self.key_repeat_delay>0 and self.key_repeat_interval>0:
+            self._key_repeat(depressed, keycode)
+
+    def _key_repeat(self, depressed, keycode):
+        if not depressed and keycode in self.keycodes_pressed:
+            """ stop the timer and clear this keycode: """
+            log.debug("key repeat: clearing timer for %s", keycode)
+            gobject.source_remove(self.keycodes_pressed[keycode])
+            del self.keycodes_pressed[keycode]
+        elif depressed and keycode not in self.keycodes_pressed:
+            """ we must ping the server regularly for as long as the key is still pressed: """
+            LATENCY_JITTER = 100
+            MIN_DELAY = 20
+            delay = max(self.key_repeat_delay-LATENCY_JITTER, MIN_DELAY)
+            interval = max(self.key_repeat_interval-LATENCY_JITTER, MIN_DELAY)
+            def continue_key_repeat(*args):
+                #if the key is still pressed (redundant check?)
+                #confirm it and continue, otherwise stop
+                log.debug("continue_key_repeat")
+                if keycode in self.keycodes_pressed:
+                    self.send(["key-repeat", keycode])
+                    return  True
+                else:
+                    del self.keycodes_pressed[keycode]
+                    return  False
+            def start_key_repeat(*args):
+                #if the key is still pressed (redundant check?)
+                #confirm it and start repeat:
+                log.debug("start_key_repeat")
+                if keycode in self.keycodes_pressed:
+                    self.send(["key-repeat", keycode])
+                    self.keycodes_pressed[keycode] = gobject.timeout_add(interval, continue_key_repeat)
+                else:
+                    del self.keycodes_pressed[keycode]
+                return  False   #never run this timer again
+            log.debug("key repeat: starting timer for %s with delay %s and interval %s", keycode, delay, start_key_repeat)
+            self.keycodes_pressed[keycode] = gobject.timeout_add(delay, start_key_repeat)
+
+    def clear_repeat(self):
+        for timer in self.keycodes_pressed.values():
+            gobject.source_remove(timer)
+        self.keycodes_pressed = {}
+
+    def query_xkbmap(self):
+        self.xkbmap_layout, self.xkbmap_variant, self.xkbmap_variants = self._client_extras.get_layout_spec()
+        self.xkbmap_print, self.xkbmap_query, self.xmodmap_data = self._client_extras.get_keymap_spec()
 
     def _keys_changed(self, *args):
         self._keymap = gtk.gdk.keymap_get_default()
@@ -518,13 +539,22 @@ class XpraClient(gobject.GObject):
 
     def _do_keys_changed(self, sendkeymap=False):
         self._keymap_changing = False
-        self._modifier_map = grok_modifier_map(gtk.gdk.display_get_default())
+        self._modifier_map = self._client_extras.grok_modifier_map(gtk.gdk.display_get_default())
         if sendkeymap:
             #old clients won't know what to do with it, but that's ok
             self.query_xkbmap()
-            log.info("keys_changed")
-            (_, _, current_mask) = gtk.gdk.get_default_root_window().get_pointer()
-            self.send(["keymap-changed", self.xkbmap_print, self.xkbmap_query, self.xmodmap_data, self.mask_to_names(current_mask)])
+            log("keys_changed")
+            if self.xkbmap_layout:
+                self.send_layout()
+            self.send_keymap()
+
+    def send_layout(self):
+        self.send(["layout-changed", nn(self.xkbmap_layout), nn(self.xkbmap_variant)])
+
+    def send_keymap(self):
+        (_, _, current_mask) = gtk.gdk.get_default_root_window().get_pointer()
+        self.send(["keymap-changed", nn(self.xkbmap_print), nn(self.xkbmap_query), nn(self.xmodmap_data), self.mask_to_names(current_mask)])
+
 
     def update_focus(self, id, gotit):
         def send_focus(_id):
@@ -534,11 +564,13 @@ class XpraClient(gobject.GObject):
                 self.send(["focus", _id, self.mask_to_names(current_mask)])
             else:
                 self.send(["focus", _id])
-
+        log.debug("update_focus(%s,%s) _focused=%s", id, gotit, self._focused)
         if gotit and self._focused is not id:
+            self.clear_repeat()
             send_focus(id)
             self._focused = id
         if not gotit and self._focused is id:
+            self.clear_repeat()
             send_focus(0)
             self._focused = None
 
@@ -560,26 +592,52 @@ class XpraClient(gobject.GObject):
             capabilities_request["challenge_response"] = hash
         if self.compression_level:
             capabilities_request["deflate"] = self.compression_level
+        if self.encoding:
+            capabilities_request["encoding"] = self.encoding
+        capabilities_request["encodings"] = ENCODINGS
         if self.jpegquality:
             capabilities_request["jpeg"] = self.jpegquality
         self.query_xkbmap()
+        if self.xkbmap_layout:
+            capabilities_request["xkbmap_layout"] = self.xkbmap_layout
+            if self.xkbmap_variant:
+                capabilities_request["xkbmap_variant"] = self.xkbmap_variant
         if self.xkbmap_print:
             capabilities_request["keymap"] = self.xkbmap_print
         if self.xkbmap_query:
             capabilities_request["xkbmap_query"] = self.xkbmap_query
         if self.xmodmap_data:
             capabilities_request["xmodmap_data"] = self.xmodmap_data
+        capabilities_request["cursors"] = True
+        capabilities_request["bell"] = True
+        capabilities_request["clipboard"] = self.clipboard_enabled
+        capabilities_request["notifications"] = self._client_extras.can_notify()
+        capabilities_request["packet_size"] = True
         (_, _, current_mask) = gtk.gdk.get_default_root_window().get_pointer()
         modifiers = self.mask_to_names(current_mask)
         log.debug("sending modifiers=%s" % str(modifiers))
         capabilities_request["modifiers"] = modifiers
         root_w, root_h = gtk.gdk.get_default_root_window().get_size()
         capabilities_request["desktop_size"] = [root_w, root_h]
+        capabilities_request["png_window_icons"] = True
+        key_repeat = self._client_extras.get_keyboard_repeat()
+        if key_repeat:
+            capabilities_request["key_repeat"] = key_repeat
         self.send(["hello", capabilities_request])
 
-    def send_jpeg_quality(self):
+    def send_jpeg_quality(self, q):
+        assert q>0 and q<100
+        self.jpegquality = q
         self.send(["jpeg-quality", self.jpegquality])
 
+    def send_refresh(self, id):
+        self.send(["buffer-refresh", id, True, 95])
+        self._refresh_requested = True
+
+    def send_refresh_all(self):
+        log.debug("Automatic refresh for all windows ")
+        self.send_refresh(-1)
+
     def _process_disconnect(self, packet):
         log.error("server requested disconnect: %s" % str(packet))
         self.quit()
@@ -607,7 +665,13 @@ class XpraClient(gobject.GObject):
 
     def _process_hello(self, packet):
         (_, capabilities) = packet
-        self._raw_keycodes_feature = capabilities.get("raw_keycodes_feature", False)
+        self.server_capabilities = capabilities
+        if not self.session_name:
+            self.session_name = capabilities.get("session_name", "Xpra")
+        import glib
+        glib.set_application_name(self.session_name)
+        self._raw_keycodes_feature = capabilities.get("raw_keycodes_feature", False) and \
+                            (self.xkbmap_print is not None or self.xkbmap_query is not None or self.xmodmap_data is not None)
         self._focus_modifiers_feature = capabilities.get("raw_keycodes_feature", False)
         if "deflate" in capabilities:
             self._protocol.enable_deflate(capabilities["deflate"])
@@ -626,10 +690,45 @@ class XpraClient(gobject.GObject):
                          "Please complain to "
                          "parti-discuss@partiwm.org"
                          % (avail_w, avail_h, root_w, root_h))
-        if self._clipboard_helper:
-            self._clipboard_helper.send_all_tokens()
-        self._client_extras.handshake_complete(capabilities)
+        self._protocol._send_size = capabilities.get("packet_size", False)
+        randr = capabilities.get("resize_screen", False)
+        log.debug("server has randr: %s" % randr)
+        if randr:
+            display = gtk.gdk.display_get_default()
+            i=0
+            while i<display.get_n_screens():
+                screen = display.get_screen(i)
+                screen.connect("size-changed", self._screen_size_changed)
+                i += 1
+        e = capabilities.get("encoding")
+        if e and e!=self.encoding:
+            log.debug("server is using %s encoding" % e)
+            self.encoding = e
+        self.bell_enabled = capabilities.get("bell", False)
+        self.notifications_enabled = capabilities.get("notifications", False)
+        clipboard_server_support = capabilities.get("clipboard", True) 
+        self.clipboard_enabled = clipboard_server_support and self._client_extras.supports_clipboard()
+        #ui may want to know this is now set:
+        self.emit("clipboard-toggled")
+        self.key_repeat_delay, self.key_repeat_interval = capabilities.get("key_repeat", (-1,-1))
         self.emit("handshake-complete")
+        if clipboard_server_support:
+            #from now on, we will send a message to the server whenever the clipboard flag changes:
+            self.connect("clipboard-toggled", self.send_clipboard_enabled_status)
+
+    def send_clipboard_enabled_status(self, *args):
+        self.send(["set-clipboard-enabled", self.clipboard_enabled])
+
+    def set_encoding(self, encoding):
+        assert encoding in ENCODINGS
+        assert encoding in self.server_capabilities.get("encodings", ["rgb24"])
+        self.encoding = encoding
+        self.send(["encoding", encoding])
+
+    def _screen_size_changed(self, *args):
+        root_w, root_h = gtk.gdk.get_default_root_window().get_size()
+        log.debug("sending updated screen size to server: %sx%s", root_w, root_h)
+        self.send(["desktop_size", root_w, root_h])
 
     def _process_new_common(self, packet, override_redirect):
         (_, id, x, y, w, h, metadata) = packet
@@ -650,6 +749,57 @@ class XpraClient(gobject.GObject):
         window = self._id_to_window[id]
         window.draw(x, y, width, height, coding, data)
 
+    def _process_cursor(self, packet):
+        (_, new_cursor) = packet
+        cursor = None
+        if len(new_cursor)>0:
+            (_, _, w, h, xhot, yhot, serial, pixels) = new_cursor
+            log.debug("new cursor at %s,%s with serial=%s, dimensions: %sx%s, len(pixels)=%s" % (xhot,yhot, serial, w,h, len(pixels)))
+            import array
+            bytes = array.array('b')
+            bytes.fromstring(pixels)
+            pixbuf = gtk.gdk.pixbuf_new_from_data(pixels, gtk.gdk.COLORSPACE_RGB, True, 8, w, h, w * 4)
+            x = max(0, min(xhot, w-1))
+            y = max(0, min(yhot, h-1))
+            size = gtk.gdk.display_get_default().get_default_cursor_size()
+            if size>0 and (size<w or size<h):
+                ratio = float(max(w,h))/size
+                pixbuf = pixbuf.scale_simple(int(w/ratio), int(h/ratio), gtk.gdk.INTERP_BILINEAR)
+                x = int(x/ratio)
+                y = int(y/ratio)
+            cursor = gtk.gdk.Cursor(gtk.gdk.display_get_default(), pixbuf, x, y)
+        for window in self._window_to_id.keys():
+            window.window.set_cursor(cursor)
+
+    def _process_bell(self, packet):
+        if not self.bell_enabled:
+            return
+        (_, id, device, percent, pitch, duration, bell_class, bell_id, bell_name) = packet
+        gdkwindow = None
+        if id!=0:
+            try:
+                gdkwindow = self._id_to_window[id].window
+            except:
+                pass
+        if gdkwindow is None:
+            gdkwindow = gtk.gdk.get_default_root_window()
+        log("_process_bell(%s) gdkwindow=%s", packet, gdkwindow)
+        self._client_extras.system_bell(gdkwindow, device, percent, pitch, duration, bell_class, bell_id, bell_name)
+
+    def _process_notify_show(self, packet):
+        if not self.notifications_enabled:
+            return
+        (_, dbus_id, id, app_name, replaces_id, app_icon, summary, body, expire_timeout) = packet
+        log("_process_notify_show(%s)", packet)
+        self._client_extras.show_notify(dbus_id, id, app_name, replaces_id, app_icon, summary, body, expire_timeout)
+
+    def _process_notify_close(self, packet):
+        if not self.notifications_enabled:
+            return
+        (_, id) = packet
+        log("_process_notify_close(%s)", id)
+        self._client_extras.close_notify(id)
+
     def _process_window_metadata(self, packet):
         (_, id, metadata) = packet
         window = self._id_to_window[id]
@@ -668,13 +818,17 @@ class XpraClient(gobject.GObject):
         if window._refresh_timer:
             gobject.source_remove(window._refresh_timer)
         window.destroy()
+        if len(self._id_to_window)==0:
+            log.debug("last window gone, clearing key repeat")
+            self.clear_repeat()
 
     def _process_connection_lost(self, packet):
         log.error("Connection lost")
         self.quit()
 
     def _process_gibberish(self, packet):
-        [_, data] = packet
+        (_, data) = packet
+        log.info("Received uninterpretable nonsense: %s", repr(data))
         self.emit("received-gibberish", data)
 
     _packet_handlers = {
@@ -684,6 +838,10 @@ class XpraClient(gobject.GObject):
         "new-window": _process_new_window,
         "new-override-redirect": _process_new_override_redirect,
         "draw": _process_draw,
+        "cursor": _process_cursor,
+        "bell": _process_bell,
+        "notify_show": _process_notify_show,
+        "notify_close": _process_notify_close,
         "window-metadata": _process_window_metadata,
         "configure-override-redirect": _process_configure_override_redirect,
         "lost-window": _process_lost_window,
@@ -696,8 +854,8 @@ class XpraClient(gobject.GObject):
         packet_type = packet[0]
         if (isinstance(packet_type, str)
             and packet_type.startswith("clipboard-")):
-            if self._clipboard_helper:
-                self._clipboard_helper.process_clipboard_packet(packet)
+            if self.clipboard_enabled:
+                self._client_extras.process_clipboard_packet(packet)
         else:
             self._packet_handlers[packet_type](self, packet)
 
