@@ -1,4 +1,5 @@
 # This file is part of Parti.
+# Copyright (C) 2011 Antoine Martin <antoine@devloop.org.uk>
 # Copyright (C) 2008, 2009, 2010 Nathaniel Smith <njs@pobox.com>
 # Parti is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -106,6 +107,7 @@ class Protocol(object):
         self.jpegquality = 0
         self._source_has_more = False
         self._recv_counter = 0
+        self._send_size = False
         self._closed = False
         self._read_decoder = IncrBDecode()
         self._compressor = None
@@ -128,19 +130,36 @@ class Protocol(object):
         if self._write_queue.empty() and self._source_has_more:
             self._flush_one_packet_into_buffer()
 
+    def _queue_write(self, data, flush=False):
+        if len(data)==0:
+            return
+        if self._compressor is None:
+            self._write_queue.put(data)
+            return
+        c = self._compressor.compress(data)
+        if c:
+            self._write_queue.put(c)
+        if not flush:
+            return
+        c = self._compressor.flush(zlib.Z_SYNC_FLUSH)
+        if c:
+            self._write_queue.put(c)
+
     def _flush_one_packet_into_buffer(self):
         if not self.source:
             return
         packet, self._source_has_more = self.source.next_packet()
         if packet is not None:
             log("writing %s", dump_packet(packet), type="raw.write")
-            data_payload = bencode(packet)
-            data = data_payload
-            if self._compressor is not None:
-                self._write_queue.put(self._compressor.compress(data))
-                self._write_queue.put(self._compressor.flush(zlib.Z_SYNC_FLUSH))
+            data = bencode(packet)
+            l = len(data)
+            if l<=1024:
+                #send size and data together (low copy overhead):
+                self._queue_write("PS%014d%s" % (l, data), True)
             else:
-                self._write_queue.put(data)
+                if self._send_size:
+                    self._queue_write("PS%014d" % l)
+                self._queue_write(data, True)
 
     def _write_thread_loop(self):
         try:
@@ -156,8 +175,7 @@ class Protocol(object):
                         log("write thread: writing %s", repr_ellipsized(buf))
                         buf = buf[self._conn.write(buf):]
                 except (OSError, IOError, socket.error), e:
-                    log.info("Error writing to connection: %s", e)
-                    main_thread_call(self._connection_lost)
+                    main_thread_call(self._connection_lost, "Error writing to connection: %s" % e)
                     break
                 except TypeError:
                     assert self._closed
@@ -175,8 +193,7 @@ class Protocol(object):
             try:
                 buf = self._conn.read(8192)
             except (ValueError, OSError, IOError, socket.error), e:
-                log.info("Error reading from connection: %s", e)
-                main_thread_call(self._connection_lost)
+                main_thread_call(self._connection_lost, "Error reading from connection: %s" % e)
                 return
             except TypeError:
                 assert self._closed
@@ -196,13 +213,15 @@ class Protocol(object):
                 break
         log("read thread: ended")
 
-    def _connection_lost(self):
-        log("_connection_lost")
+    def _connection_lost(self, message="", exc_info=False):
+        log.info("connection lost: %s", message, exc_info=exc_info)
         if not self._closed:
             self._process_packet_cb(self, [Protocol.CONNECTION_LOST])
             self.close()
 
     def _handle_read(self):
+        if self._closed:
+            return
         log("main thread: woken to handle read data")
         while not self._closed:
             self._read_loop_count.inc()
@@ -213,15 +232,14 @@ class Protocol(object):
                 log("main thread: processed all read data")
                 return
             if not buf:
-                self._connection_lost()
+                self._connection_lost("empty marker in read queue")
                 return
             if self._decompressor is not None:
                 buf = self._decompressor.decompress(buf)
             try:
                 self._read_decoder.add(buf)
             except:
-                log.error("bencoder read buffer is in an inconsistent state, cannot continue")
-                self._connection_lost()
+                self._connection_lost("read buffer is in an inconsistent state, cannot continue", exc_info=True)
                 return
             while not self._closed:
                 had_deflate = (self._decompressor is not None)
@@ -229,10 +247,9 @@ class Protocol(object):
                     result = self._read_decoder.process()
                 except ValueError:
                     # Peek at the data we got, in case we can make sense of it:
-                    self._process_packet([Protocol.GIBBERISH,
-                                          self._read_decoder.unprocessed()])
+                    self._process_packet([Protocol.GIBBERISH, self._read_decoder.unprocessed()])
                     # Then hang up:
-                    self._connection_lost()
+                    self._connection_lost("gibberish received")
                     return
                 if result is None:
                     break
