@@ -1,4 +1,5 @@
 # This file is part of Parti.
+# Copyright (C) 2010, 2011 Antoine Martin <antoine@devloop.org.uk>
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
 # Parti is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
@@ -62,7 +63,7 @@ class ChildReaper(object):
     def __call__(self, signum, frame):
         while 1:
             try:
-                pid, status = os.waitpid(-1, os.WNOHANG)
+                pid, _ = os.waitpid(-1, os.WNOHANG)
             except OSError:
                 break
             if pid == 0:
@@ -249,6 +250,7 @@ def run_server(parser, opts, mode, xpra_file, extra_args):
                                      executable=xvfb_executable)
         except OSError, e:
             sys.stderr.write("Error starting Xvfb: %s\n" % (e,))
+            return
         raw_cookie = os.urandom(16)
         baked_cookie = raw_cookie.encode("hex")
         try:
@@ -297,7 +299,8 @@ def run_server(parser, opts, mode, xpra_file, extra_args):
 
     # This import is delayed because the module depends on gtk:
     import xpra.server
-    app = xpra.server.XpraServer(upgrading, sockets, opts.password_file, opts.pulseaudio, opts.clipboard, opts.randr)
+    app = xpra.server.XpraServer(upgrading, sockets, opts.session_name,
+                                 opts.password_file, opts.pulseaudio, opts.clipboard, opts.randr, opts.encoding)
     def cleanup_socket(self):
         print "removing socket"
         try:
